(**************************************************************************)
(*                                                                        *)
(*  This file is part of the Frama-C's E-ACSL plug-in.                    *)
(*                                                                        *)
(*  Copyright (C) 2012-2025                                               *)
(*    CEA (Commissariat à l'énergie atomique et aux énergies              *)
(*         alternatives)                                                  *)
(*                                                                        *)
(*  you can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version 2.1                 *)
(*  for more details (enclosed in the file licenses/LGPLv2.1).            *)
(*                                                                        *)
(**************************************************************************)

open Cil_types
open Cil_datatype

let dkey = Options.Dkey.logic_normalizer

module Id_predicate =
  Datatype.Make_with_collections
    (struct
      include Predicate
      let name = "E_ACSL.Id_predicate"
      (* The function compare should never be used since we only use this
         datatype for hashtables *)
      let compare _ _ = assert false
      let equal (p1:predicate) p2 = p1 == p2
      let structural_descr = Structural_descr.t_abstract
      let hash = Logic_utils.hash_predicate
      let rehash = Datatype.identity
      let mem_project = Datatype.never_any_project
    end)

(** Functor for creating memoization modules for the normalization of
    predicates, terms, and logic functions. They are bidirectional so that the
    original version of any such object can be obtained. This is useful for two
    applications: 1. better user feedback - by looking up the original form we
    avoid confronting the user with internally generated names. 2. from the
    original [logic_info]s we need to be able to enumerate all the generated
    [logic_info]s in order to add their function declarations to the C file. *)
module Memo (H : Hashtbl.S) = struct
  module M = Bimap.Make (H)

  let clear = M.clear

  let derived = M.tail_or_self
  let derived_opt = M.tail_opt
  let original = M.head_or_self
  let original_opt = M.head_opt

  let memoize process x =
    try Some (M.tail x) with
    | Not_found ->
      match process x with
      | Some y -> M.add x y; Some y
      | None -> None
end

(** memoization of normalised predicates *)
module Predicates = Memo (Id_predicate.Hashtbl)

(** memoization of normalised terms *)
module Terms = Memo (Misc.Id_term.Hashtbl)

(** memoization of here-specialised logic definitions *)
module Here_specialized = Memo (Logic_info.Hashtbl)

(** memoization of auxiliary logic functions generated by inductive extraction *)
module Auxiliary_functions = Memo (Logic_info.Hashtbl)

(** memoization of logic_infos extracted from inductive definition *)
module Inductive_extractions = struct
  include Memo (Logic_info.Hashtbl)

  let extract_predicate = memoize @@ fun li ->
    try Some (Inductive.extract_predicate li)
    with Inductive.Unsupported ->
      Options.warning
        "Unable to extract executable code from inductive definition of %a.@ \
         See language reference manual for details on the supported forms."
        Printer.pp_logic_info li;
      Some li
end

let pp_logic_info fmt li =
  Printer.pp_global_annotation fmt (Dfun_or_pred (li, Location.unknown))

(** Retrieve information about generated [logic_info]s and the original forms
    from which they stem. *)
module Logic_infos : sig
  val origin_of_lv : logic_var -> logic_var
  val generated_of : logic_info -> logic_info list
end = struct
  let original_of_lv lv =
    let li = {l_var_info = lv;
              l_labels = [];
              l_tparams = [];
              l_type = None;
              l_profile = [];
              l_body = LBnone} in
    match Here_specialized.original_opt li with
    | Some orig -> Some orig
    | None -> match Auxiliary_functions.original_opt li with
      | Some orig -> Some orig
      | None -> Inductive_extractions.original_opt li

  let rec origin_of_lv lv =
    match original_of_lv lv with
    | None -> lv
    | Some origin -> origin_of_lv origin.l_var_info

  let generated_of li =
    List.filter (fun li' -> not @@ Logic_info.equal li li') @@
    Option.to_list (Inductive_extractions.derived_opt li) @
    Option.to_list (Here_specialized.derived_opt li) @
    Auxiliary_functions.M.tails li
end

(** E-ACSL currently is not in general capable of translationg predicates and
    logic functions with labels. However often such functions are used
    while supplying the Here label.
    In this case we can specialize the logic function, such that it has no
    labels, substituting all occurrences of the original labels by the Here
    label. Then the specialized function can be used instead of the original
    one whenever all labels are Here. *)
module Here_inliner : sig
  val preprocess_pred : predicate -> predicate option
  val preprocess_term : term -> term option

  (** apply a preprocessor to the result of another *)
  val bind : ('a -> 'a option) -> 'a -> 'a option -> 'a option
end = struct

  let here = BuiltinLabel Here
  let is_here = (=) here
  let are_all_here labels = List.length labels > 0 && List.for_all is_here labels

  (** substitute occurrences of li_old by li_new; also substitute occurrences
      of li_old's formal parameters by li_new's parameters; li_new does not
      have any labels so substitute occurrences of li_old's labels by Here *)
  let substitute li_old li_new =
    let formal_substitutions =
      Logic_var.Map.of_seq
        (List.to_seq @@ List.combine li_old.l_profile li_new.l_profile)
    in
    let is_obsolete_label =
      let obsolete_labels = Logic_label.Set.of_list @@ here :: li_old.l_labels in
      fun l -> Logic_label.Set.mem l obsolete_labels
    in
    object
      inherit Visitor.frama_c_inplace

      method! vlogic_var_use lv =
        match Logic_var.Map.find_opt lv formal_substitutions with
        | None -> DoChildren
        | Some lv' -> ChangeTo lv'

      method !vpredicate p =
        match p.pred_content with
        | Papp (li, labels, args) when
            Logic_info.equal li li_old && List.for_all is_obsolete_label labels ->
          let p = {p with pred_content = Papp (li_new, [], args)} in
          ChangeDoChildrenPost (p, fun x -> x)
        | _ -> DoChildren

      method !vterm t =
        match t.term_node with
        | Tat (body, label) when is_obsolete_label label ->
          ChangeDoChildrenPost (body, fun x -> x)
        | Tapp (li, labels, args) when
            Logic_info.equal li li_old && List.for_all is_obsolete_label labels ->
          let t = {t with term_node = Tapp (li_new, [], args)} in
          ChangeDoChildrenPost (t, fun x -> x)
        | _ -> DoChildren

      method !vlogic_info_use l =
        if Logic_info.equal l li_old
        then ChangeTo li_new
        else DoChildren

      method !vlogic_label l =
        if is_obsolete_label l then ChangeTo here else DoChildren
    end

  let specialize li =
    let f li =
      let lv_name = Functions.RTL.mk_gen_name (li.l_var_info.lv_name ^ "_here") in
      let vi = {li.l_var_info with lv_name; lv_id = Cil_const.new_raw_id ()} in
      let li' = {
        li with
        l_var_info = vi;
        l_labels = [];
        l_profile =
          List.map (fun lv -> {lv with lv_id = Cil_const.new_raw_id ()}) li.l_profile
      } in
      let li' = Visitor.visitFramacLogicInfo (substitute li li') li' in
      let pred_or_term =
        match li.l_type with None -> "predicate" | Some _ -> "logic function"
      in
      Options.feedback ~dkey ~level:2
        "specializing %s %a for use cases using the Here label"
        pred_or_term
        Printer.pp_logic_info li;
      Options.feedback ~dkey ~level:3
        "specialized version of %s %a:@ @[%a@]"
        pred_or_term
        Printer.pp_logic_info li
        pp_logic_info li';
      Some li'
    in
    Option.get @@ Here_specialized.memoize f li

  let inliner = object (self)
    inherit Visitor.frama_c_inplace

    method !vpredicate p =
      match p.pred_content with
      | Papp (li, labels, args) when are_all_here labels ->
        Options.feedback ~dkey ~level:3
          "inlining Here labels for predicate: @[%a@]"
          Printer.pp_predicate p;
        let li = specialize li in
        (* The Cil visitor does not descend into [logic_info] definitions. *)
        let li = Visitor.visitFramacLogicInfo self li in
        let p = {p with pred_content = Papp(li, [], args)} in
        ChangeDoChildrenPost (p, fun x -> x)
      | _ -> DoChildren

    method !vterm t =
      match t.term_node with
      | Tapp(li, labels, args) when are_all_here labels ->
        Options.feedback ~dkey ~level:3
          "inlining Here labels for term: %a"
          Printer.pp_term t;
        let li = specialize li in
        (* The Cil visitor does not descend into [logic_info] definitions. *)
        let li = Visitor.visitFramacLogicInfo self li in
        let t = Logic_const.term ~loc:t.term_loc (Tapp (li, [], args)) t.term_type in
        ChangeDoChildrenPost (t, fun x -> x)
      | _ -> DoChildren
  end

  let preprocess_pred p =
    match p.pred_content with
    | Papp(_li, labels, _args) when are_all_here labels ->
      Some (Visitor.visitFramacPredicate inliner p)
    | _ -> None

  let preprocess_term t =
    match t.term_node with
    | Tapp(_li, labels, _args) when are_all_here labels ->
      Some (Visitor.visitFramacTerm inliner t)
    | _ -> None

  (* apply to the result of a first preprocessor a second one *)
  let bind f orig = function
    | None -> f orig (* first preprocessor yields nothing, use original value *)
    | Some p -> match f p with
      | None -> Some p (* second preprocessor yields nothing *)
      | Some p -> Some p
end

let preprocess_pred ~loc p =
  Here_inliner.(bind preprocess_pred) p @@
  match p.pred_content with
  | Papp(li , labels, args) when Inductive.is_inductive li ->
    Inductive_extractions.extract_predicate li |> Option.map @@ fun li ->
    {p with pred_content = Papp (li, labels, args)}
  | Pvalid_read(BuiltinLabel Here as llabel, t)
  | Pvalid(BuiltinLabel Here as llabel, t)
  | Pobject_pointer(BuiltinLabel Here as llabel, t) -> begin
      match t.term_node, t.term_type with
      | TLval tlv, lty ->
        let init =
          Logic_const.pinitialized
            ~loc
            (llabel, Logic_utils.mk_logic_AddrOf ~loc tlv lty)
        in
        (* need to store a copy, to avoid p to appear in its own preprocessed
           form (otherwise it loops) *)
        let p_copy =
          match p.pred_content with
          | Pvalid_read _ -> Logic_const.pvalid_read ~loc (llabel, t)
          | Pvalid _ -> Logic_const.pvalid ~loc (llabel, t)
          | Pobject_pointer _ -> Logic_const.pobject_pointer ~loc (llabel, t)
          | _ -> assert false
        in
        Some (Logic_const.pand ~loc (init, p_copy))
      | _ -> None
    end
  | _ -> None

let preprocess_term ~loc t =
  Here_inliner.(bind preprocess_term) t @@
  match t.term_node with
  | Tapp(li, lst, [ t1; t2; {term_node = Tlambda([ k ], predicate)}])
    when li.l_body = LBnone && li.l_var_info.lv_name = "\\numof" ->
    let logic_info = Cil_const.make_logic_info "\\sum" in
    logic_info.l_type <- li.l_type;
    logic_info.l_tparams <- li.l_tparams;
    logic_info.l_labels <- li.l_labels;
    logic_info.l_profile <- li.l_profile;
    logic_info.l_body <- li.l_body;
    let conditional_term =
      Logic_const.term ~loc
        (Tif(predicate, Cil.lone (), Cil.lzero ())) Linteger
    in
    let lambda_term =
      Logic_const.term ~loc (Tlambda([ k ], conditional_term)) Linteger
    in
    Some (Logic_const.term ~loc
            (Tapp(logic_info, lst, [ t1; t2; lambda_term ])) Linteger)
  | Tapp(li , labels, args) when Inductive.is_inductive li ->
    Inductive_extractions.extract_predicate li |> Option.map @@ fun li ->
    (Logic_const.term ~loc:t.term_loc (Tapp (li, labels, args)) t.term_type)
  | _ -> None

let preprocessor = object
  inherit E_acsl_visitor.visitor dkey

  method !vannotation annot =
    match annot with
    | Dfun_or_pred _ -> Cil.DoChildren
    | _ -> Cil.SkipChildren

  method !vpredicate p =
    let loc = p.pred_loc in
    ignore @@ Predicates.memoize (fun p ->
        begin match preprocess_pred ~loc p with
          | Some pot ->
            Options.debug ~dkey ~level:4
              "@[memoised predicate@;<2>@[%a@]@;<2>as@;<2>@[%a@]@]"
              Printer.pp_predicate p
              Printer.pp_predicate pot;
            Some pot
          | None -> None
        end
      ) p;
    Cil.DoChildren

  method !vterm t =
    let loc = t.term_loc in
    ignore @@ Terms.memoize (fun t ->
        begin match preprocess_term ~loc t with
          | Some term ->
            Options.debug ~dkey ~level:4
              "@[memoised term@;<2>@[%a@]@;<2>as@;<2>@[%a@]@]"
              Printer.pp_term t
              Printer.pp_term term;
            Some term
          | None -> None
        end
      ) t;
    Cil.DoChildren
end

let preprocess ast =
  preprocessor#visit_file ast;
  Inductive.Derived_functions.iter Auxiliary_functions.M.add

let preprocess_annot annot =
  ignore @@ preprocessor#visit_code_annot annot

let preprocess_predicate p =
  ignore @@ preprocessor#visit_predicate p

let get_pred = Predicates.derived
let get_orig_pred = Predicates.original
let get_term = Terms.derived
let get_orig_term = Terms.original

let clear () =
  Terms.clear ();
  Predicates.clear ();
  Auxiliary_functions.clear ();
  Inductive_extractions.clear ();
  Inductive.clear ();
  Here_specialized.clear ()
